#! --double #--nothing 

# Version 0.9.5, August 2006
# Adds rotational acceleration in friction term, Voellmy rheology
# and earth-pressure coefficient K

# Script for 2D transport of material on a DEM, typically for routing of
# debris-flows originating from a landslide. The model uses a simplified,
# depth integrated, form of Bingham rheology to calculate the relation
# between stress and velocity. The numerical solution (Lax scheme) is first
# order accurate in time, so the error would be O(dT, dX).

# Body thickness, pressures and velocities are calculated in the XYZ plane,
# and not paralell/perpendicular to the slope. Directions in the XY plane
# are considered positive towards the East and the North -hence gradients
# are negative to the E and N, and velocities are positive to the E and the N-.

# IMPORTANT NOTICE: This is a developing version of the model, and it is
# provided as is, for illustration and learning purposes. Other uses of the
# model are not allowed at this stage.
#
# 2006 Santiago Beguera-Portugus. Some rights reserved.
# email: s.begueria@geo.uu.nl, sbegueria@ipe.csic.es

# This script requires a working instalation of PCRaster spatial modelling
# software (http://pcraster.geo.uu.nl). Run it with:
#
# pcrcalc -1 -f DFMOV_v5.mod [simulation_id] [gamma] [coh] [visco] 
# [friction / Chzy roughness] [timestep] [n_timesteps]
#
# where: [simulation_id] is a string which identifies the current simulation
#(it is used to give names to the reported results); [gamma] is typically
# around 15-20; [coh] is typically O(1); [visco] is typically O(1E-1); [friction]
# is typically O(1E-1); [timestep] is the duration of one timestep (seconds) 
# [n_timesteps] is the total number of timesteps of the simulation, hence the
# total simulation time in seconds equals [timestep] * [n_timesteps]; these two
# parameters vary greatly from one simulation to another.


binding

 # Input maps
 DEM          = dem.map;           # Elevation map
 H_ini        = h_ini.map;         # Initial body thickness (m, vertical)
 UH_ini       = uh_ini.map;        # Initial body momentum, x direction (m2.s-1)
 VH_ini       = vh_ini.map;        # Initial body momentum, y direction (m2.s-1)
 Inlet        = inlet.map;         # Domain inlet (upper boundary)
 Outlet       = outlet.map;        # Domain outlet (bottom boundary)
 Channels     = channels.map;      # Channels (channel depth, m)
# Err_weights  = errw.map;          # Observed deposition (error weights, scalar)
 SamplePnts   = points.map;        # Map with sampling points (nominal)

 # Timeseries (inlet boundary condition)
 InputH       = inp_h.tss;         # Input sedigraph (H at inlet, m)

 # Timestep control
 dT           = scalar($6);        # Timeslice (seconds)
 Timesteps    = scalar($7);        # Number of timesteps of the simulation
 MinNLoops    = 1;                 # Minimum number of internal loops
 MaxNLoops    = 1024;              # Maximum number of internal loops
 InitialLoops = 1;

 # Other constants
 g = 9.8;                          # gravity acceleration (m.s-2)
 verysmall = 0.1;                  # threshold to avoid dividing by very small values

 # Mud rheology
 Gamma        = scalar($2);        # Bulk unit weight of df (kPa.m-1)
 Cohesion     = scalar($3);        # Cohesivity (yield strenght due to cohesion) (kPa)
 Visco        = scalar($4);        # Dynamic viscosity (kPa.s)
 Friction     = scalar($5);        # Angle of internal friction ()
 #Chezy        = scalar($6);        # Chzy roughness coefficient (m.s-2)
 #Friction_v   = scalar($7);        # Frict. coeff. 4 Voellmy, typically << Friction ()


areamap

 DEM;


timer

 ## ---- CHANGES ARE ALLOWED HERE --------------------------------------------

 1 Timesteps 1;                    # repetitions: from t = A to t = B, incr. = 1
 rep_a = 1, 1+1..endtime;          # reporting intervals (at t=1, then each
 rep_b = 1, 10+10..endtime;        # n+n until endtime)
 end = endtime;
 
 ## ---- END OF ALLOWED CHANGES ----------------------------------------------

initial

 ## Fixed terms (constants)
 Sqrt2 = sqrt(2);                                 # just the square root of two 
 TanDelta = tan(Friction);                        # tangens of the friction ang.
 #TanDelta_v = tan(Friction_v);                    # tangens of the friction ang.
 CL = celllength();                               # cell length
 CA = CL*CL;                                      # cell area
 Inlet_length = CL * maptotal(                    # length of the inlet area (m)
   if(Inlet, scalar(1)));

 ## Fixed maps
 # calculation domain
 Clone = if (defined(DEM), boolean(1));

 # determine domain boundaries
 l = shift0(Clone, 0,-1);                         # left cell
 r = shift0(Clone, 0, 1);                         # right cell
 u = shift0(Clone,-1, 0);                         # up cell
 d = shift0(Clone, 1, 0);                         # down cell
 ldom = if(Clone and l, boolean(1), boolean(0));  # TRUE: left cell is inside
 rdom = if(Clone and r, boolean(1), boolean(0));  # the DTM domain
 udom = if(Clone and u, boolean(1), boolean(0));  # FALSE: left cell is outside
 ddom = if(Clone and d, boolean(1), boolean(0));  # the DTM domain

 # determine channel boundaries
 IsChannel = if(Channels gt 0, boolean(1), boolean(0));
 l = shift0(IsChannel, 0,-1);
 r = shift0(IsChannel, 0, 1);
 u = shift0(IsChannel,-1, 0);
 d = shift0(IsChannel, 1, 0);
 lchan = if(Clone and l, boolean(1), boolean(0)); # TRUE: left cell is channel
 rchan = if(Clone and r, boolean(1), boolean(0));  
 uchan = if(Clone and u, boolean(1), boolean(0));  
 dchan = if(Clone and d, boolean(1), boolean(0));  

 # intercell lenght for central differences (dx, dy)
 dx = if(rdom and ldom, 2*CL, CL);
 dy = if(udom and ddom, 2*CL, CL);

 # basal surface gradient and vertical curvature; 
 # central diffs. within the DEM domain, upwind diffs. at the boundaries
 Z = DEM;
 # Z of neighbours, fwd. rule for boundaries
 Z_l = if(ldom, shift0(Z, 0,-1), Z);              # value of cell to the left
 Z_r = if(rdom, shift0(Z, 0, 1), Z);              # right
 Z_u = if(udom, shift0(Z,-1, 0), Z);              # up
 Z_d = if(ddom, shift0(Z, 1, 0), Z);              # down
 # Z gradient in X and Y directions
 dZ\dx = (Z_r-Z_l) / dx;                          # gradient in x: <0 to the E
 dZ\dy = (Z_u-Z_d) / dy;                          # gradient in y: <0 to the N
 # basal surface vertical curvature
 d2Z\dx2 = (Z_r-2*Z+Z_l) / (dx*dx);
 d2Z\dy2 = (Z_u-2*Z+Z_d) / (dy*dy);    
 # slope angles, negative towards E and N
 CosAlfa_x = cos(atan(dZ\dx));
 CosAlfa_y = cos(atan(dZ\dy));
 SinCosAlfa_x = sin(atan(dZ\dx)) * CosAlfa_x;
 SinCosAlfa_y = sin(atan(dZ\dy)) * CosAlfa_y;

 # earth pressure coefficient (Koch, 1998)
 SinPhi = sin(Friction);
 K_act = (1-SinPhi) / (1+SinPhi);
 K_pas = (1+SinPhi) / (1-SinPhi);

 ## Initial values
 H = H_ini;                                       # body thickness
 UH = UH_ini;                                     # momentum, x direction
 VH = VH_ini;                                     # momentum, y direction
 H_loop = H;                                      # H, for internal loop
 UH_loop = UH;                                    # ''
 VH_loop = VH;                                    # ''
 H_inlt = 0;                                      # H at the ghost inlet cells
 Vol_expect = maptotal(H*CA);                     # initial volume (m3)
 Loops = InitialLoops;                            # number of internal loops


dynamic

 # thickness and velocity at the inlet
 H_inlt = mapmaximum(timeinputscalar(InputH, Inlet));
 H_inlt = if(H_inlt lt verysmall, 0, H_inlt);
 V_inlt = -H_inlt/Visco * (Cohesion + Gamma*H_inlt*TanDelta);
#   -2 * H_inlt/Visco * (Cohesion + Gamma*H_inlt*TanDelta);
#   -H_inlt**2 / (2*Visco) * (Cohesion + Gamma * H_inlt * TanDelta), 0);

 repeat {        # main loop: only exits when the whole timestep
                 # (dT) has been calculated without errors

   # initialize variables in the internal loop
   n = 0;
   CFLmax = 0;
   Vel_aver = 0;
   H_loop = H;
   UH_loop = UH;
   VH_loop = VH;
   Vol_exp_loop = Vol_expect;
   Incr_volexploop = H_inlt * Inlet_length * abs(V_inlt) * dT/Loops;

   repeat {      # internal loop: subdivide dT in a variable number
                 # of smaller timesteps to ensure stability

     ### 0. CALCULATE NEEDED GRADIENTS

     ## values of neighbouring cells
     # H neighbours
     H_l = if(ldom, shift0(H_loop, 0,-1), H_loop);
     H_r = if(rdom, shift0(H_loop, 0, 1), H_loop);
     H_u = if(udom, shift0(H_loop,-1, 0), H_loop);
     H_d = if(ddom, shift0(H_loop, 1, 0), H_loop);
     H_u  = if(Inlet,  H_inlt, H_u);
     H_d  = if(Outlet, H_loop, H_d);
     # UH neighbours
     H_chann = max(H_loop-Channels, 0);
     FractExc = H_chann/H_loop;           # Fraction of d.f. exceeding channel
     UH_l  = if(lchan and not IsChannel, shift0(UH_loop*FractExc, 0,-1),
       shift0(UH_loop, 0,-1));
     UH_l  = if(ldom, UH_l, UH_loop);
     UH_r  = if(rchan and not IsChannel, shift0(UH_loop*FractExc, 0, 1),
       shift0(UH_loop, 0, 1));
     UH_r  = if(rdom, UH_r, UH_loop);
     UH_u  = if(uchan and not IsChannel, shift0(UH_loop*FractExc,-1, 0),
       shift0(UH_loop,-1, 0));
     UH_u  = if(udom, UH_u, UH_loop);
     UH_d  = if(dchan and not IsChannel, shift0(UH_loop*FractExc, 1, 0),
       shift0(UH_loop, 1, 0));
     UH_d  = if(ddom, UH_d, UH_loop);
     # VH neighbours
     VH_l  = if(lchan and not IsChannel, shift0(VH_loop*FractExc, 0,-1),
       shift0(VH_loop, 0,-1));
     VH_l  = if(ldom, VH_l, VH_loop);
     VH_r  = if(rchan and not IsChannel, shift0(VH_loop*FractExc, 0, 1),
       shift0(VH_loop, 0, 1));
     VH_r  = if(rdom, VH_r, VH_loop);
     VH_u  = if(uchan and not IsChannel, shift0(VH_loop*FractExc,-1, 0),
       shift0(VH_loop,-1, 0));
     VH_u  = if(udom, VH_u, UH_loop);
     VH_d  = if(dchan and not IsChannel, shift0(VH_loop*FractExc, 1, 0),
       shift0(VH_loop, 1, 0));
     VH_d  = if(ddom, VH_d, UH_loop);
     VH_u = if(Inlet,  V_inlt*H_inlt, VH_u);
     VH_d = if(Outlet, VH_loop, VH_d);
     VH_u = if(Inlet,  V_inlt*H_inlt, VH_u);
     VH_d = if(Outlet, VH_loop, VH_d);

     ## spatial averages of UH and VH
     UH_loop_a = if(H_loop eq 0, 0, 4/8*UH_loop + 1/8*(UH_l+UH_r+UH_u+UH_d));
     VH_loop_a = if(H_loop eq 0, 0, 4/8*VH_loop + 1/8*(VH_l+VH_r+VH_u+VH_d));

     ## spatial derivatives
     # gradient of the free surface (z+h)
     dZH\dx = ((Z_r + if(rdom, shift0(H_loop, 0, 1), 0)) -
       (Z_l + if(ldom, shift0(H_loop, 0,-1), 0))) / dx;
     dZH\dx = if(dZH\dx lt 0 and H_l eq 0, 0, dZH\dx); # correction for df
     dZH\dx = if(dZH\dx gt 0 and H_r eq 0, 0, dZH\dx); # borders (no pressure
     dZH\dy = ((Z_u + if(udom, shift0(H_loop,-1, 0), 0)) -
       (Z_d + if(ddom, shift0(H_loop, 1, 0), 0))) / dy;
     dZH\dy = if(dZH\dy lt 0 and H_d eq 0, 0, dZH\dy); # from outside the
     dZH\dy = if(dZH\dy gt 0 and H_u eq 0, 0, dZH\dy); # df boundaries)
     # grad U, grad V; positive -> divergence, negative -> convergence
     dU = if(H_r gt verysmall, UH_r/H_r, 0) - if(H_l gt verysmall, UH_l/H_l, 0);
     dV = if(H_u gt verysmall, VH_u/H_u, 0) - if(H_d gt verysmall, VH_d/H_d, 0);
     # grad(HU^2), grad(HV^2)
     dHU2\dx = 1/dx * (if(H_r gt verysmall, UH_r*abs(UH_r)/H_r, 0) -
       if(H_l gt verysmall, UH_l*abs(UH_l)/H_l, 0));
     dHV2\dy = 1/dy * (if(H_u gt verysmall, VH_u*abs(VH_u)/H_u, 0) -
       if(H_d gt verysmall, VH_d*abs(VH_d)/H_d, 0));
     # grad(HUV), grad(HVU)
     dHUV\dx = 1/dx * (if(H_r gt verysmall, UH_r*abs(VH_r)/H_r, 0)
       - if(H_l gt verysmall, UH_l*abs(VH_l)/H_l, 0));
     dHVU\dy = 1/dy * (if(H_u gt verysmall, VH_u*abs(UH_u)/H_u, 0)
       - if(H_d gt verysmall, VH_d*abs(UH_d)/H_d, 0));

     # earth pressure ratio: active state = dilatation; pasive st. = compression
     K_x = if (dU ge 0, K_act, K_pas); 
     K_y = if (dV ge 0, K_act, K_pas); 


     ### 1. DETERMINE FLUXES AT t = t + 1/2dt
     # dhu/dt = -d(hu2)/dx - d(huv)/dy + gF_x/gamma   #   ur/ |\
     # dhv/dt = -d(hv2)/dy - d(hvu)/dx + gF_y/gamma   #  s/   |  \ b = a cos(ab)

     ## calculate acceleration
     # all vectors, parallel to XY plane

     # GP, gravitational and pressure acceleration (m2.s-2)
     W = -g * H_loop;
     G_x = W * SinCosAlfa_x;
     G_y = W * SinCosAlfa_y;
     P_x = W * K_x*dZH\dx;
     P_y = W * K_y*dZH\dy;
     GP_x = G_x + P_x;
     GP_y = G_y + P_y;

     # I, convective acceleration (m2.s-2)
     I_x = -dHU2\dx - dHVU\dy;
     I_y = -dHV2\dy - dHUV\dx;

     # T, basal resistance (m2.s-2)
     H_sqr = H_loop*H_loop;
     # Coulomb yield stress
     Yield_x = Cohesion + Gamma*H_loop * (CosAlfa_x + d2Z\dx2/g) * TanDelta;
     Yield_y = Cohesion + Gamma*H_loop * (CosAlfa_y + d2Z\dx2/g) * TanDelta;
     # Bingham simplified rheology (Coussot, 1997)
     T_x = g/Gamma * 1.5 * (Yield_x +
       if(H_loop lt verysmall, 10, 2*Visco * abs(UH_loop) / (H_sqr)));
     T_y = g/Gamma * 1.5 * (Yield_y +
       if(H_loop lt verysmall, 10, 2*Visco * abs(VH_loop) / (H_sqr)));
     # Voellmy rheology
     #Yield_x = Gamma*H_loop * (CosAlfa_x + d2Z\dx2/g) * TanDelta_v;
     #Yield_y = Gamma*H_loop * (CosAlfa_y + d2Z\dx2/g) * TanDelta_v;
     #T_x = g * (Yield_x/Gamma + if(H_loop lt verysmall, 10,
     #  (UH_loop*UH_loop)/(H_sqr*Chezy)));
     #T_y = g * (Yield_y/Gamma + if(H_loop lt verysmall, 10,
     #  (VH_loop*VH_loop)/(H_sqr*Chezy)));

     ## estimate fluxes at t = t+(1/2)dT; Lax numerical scheme
     # UH
     UH_loop_b = if(UH_loop_a gt 0,
       max(0, UH_loop_a + 1/2*dT/Loops * (GP_x + I_x - T_x)),
       if(UH_loop_a lt 0,
       min(UH_loop_a + 1/2*dT/Loops * (GP_x - I_x + T_x), 0), 0));
     UH_loop_b = if(UH_loop_a eq 0 and GP_x gt 0,
       max(0, 1/2*dT/Loops * (GP_x + I_x - T_x)),
       if(UH_loop_a eq 0 and GP_x lt 0,
       min(1/2*dT/Loops * (GP_x - I_x + T_x), 0), UH_loop_b));
     # VH
     VH_loop_b = if(VH_loop_a gt 0,
       max(0, VH_loop_a + 1/2*dT/Loops * (GP_y + I_y - T_y)),
       if(VH_loop_a lt 0,
       min(VH_loop_a + 1/2*dT/Loops * (GP_y - I_y + T_y), 0), 0));
     VH_loop_b = if(VH_loop_a eq 0 and GP_y gt 0,
       max(0, 1/2*dT/Loops * (GP_y + I_y - T_y)),
       if(VH_loop_a eq 0 and GP_y lt 0,
       min(1/2*dT/Loops * (GP_y - I_y + T_y), 0), VH_loop_b));

     # T at time t = t+(1/2)dT (m2.s-2)
     # Bingham simplified rheology
     T_x_b = g/Gamma * 1.5 * (Yield_x +
       if(H_loop lt verysmall, 10, 2*Visco * abs(UH_loop_b) / (H_loop**2)));
     T_y_b = g/Gamma * 1.5 * (Yield_y +
       if(H_loop lt verysmall, 10, 2*Visco * abs(VH_loop_b) / (H_loop**2)));
     # Voellmy
     #T_x_b = g * (Yield_x/Gamma + if(H_loop lt verysmall, 10,
     #  (UH_loop_b*UH_loop_b)/(H_sqr*Chezy)));
     #T_y_b = g * (Yield_y/Gamma + if(H_loop lt verysmall, 10,
     #  (VH_loop_b*VH_loop_b)/(H_sqr*Chezy)));
       
     ## new fluxes at t=t+dT; Lax numerical scheme
     # UH
     UH_loop = if(UH_loop_a gt 0,
       max(0, UH_loop_a + dT/Loops * (GP_x + I_x - T_x_b)),
       if(UH_loop_a lt 0,
       min(UH_loop_a + dT/Loops * (GP_x - I_x + T_x_b), 0),
       0));
     UH_loop = if(UH_loop_a eq 0 and GP_x gt 0,
       max(0, dT/Loops * (GP_x + I_x - T_x_b)),
       if(UH_loop_a eq 0 and GP_x lt 0,
       min(dT/Loops * (GP_x - I_x + T_x_b), 0), UH_loop));
     # VH
     VH_loop = if(VH_loop_a gt 0,
       max(0, VH_loop_a + dT/Loops * (GP_y + I_y - T_y_b)),
       if(VH_loop_a lt 0,
       min(VH_loop_a + dT/Loops * (GP_y - I_y + T_y_b), 0),
       0));
     VH_loop = if(VH_loop_a eq 0 and GP_y gt 0,
       max(0, dT/Loops * (GP_y + I_y - T_y_b)),
       if(VH_loop_a eq 0 and GP_y lt 0,
       min(dT/Loops * (GP_y - I_y + T_y_b), 0), VH_loop));


     ### 2. DETERMINE D.F. THICKNESS AT t = t + dt
     # dh/dt + d(uh)/dx + d(vh)/dy = 0 -> dh = -dt * (d(uh)/dx + d(vh)/dy)
     
     ## spatial derivatives, accounting for channels
     # H neighbours
     H_l  = if(lchan and not IsChannel, shift0(H_chann, 0,-1), H_l);
     H_r  = if(rchan and not IsChannel, shift0(H_chann, 0, 1), H_r);
     H_u  = if(uchan and not IsChannel, shift0(H_chann,-1, 0), H_u);
     H_d  = if(dchan and not IsChannel, shift0(H_chann, 1, 0), H_d);
     # UH neighbours
     FractExc = H_chann/H_loop;           # Fraction of d.f. exceeding channel
     UH_l  = if(lchan and not IsChannel, shift0(UH_loop*FractExc, 0,-1),
       shift0(UH_loop, 0,-1));
     UH_l  = if(ldom, UH_l, 0);
     UH_r  = if(rchan and not IsChannel, shift0(UH_loop*FractExc, 0, 1),
       shift0(UH_loop, 0, 1));
     UH_r  = if(rdom, UH_r, 0);
     # VH neighbours
     VH_u  = if(uchan and not IsChannel, shift0(VH_loop*FractExc,-1, 0),
       shift0(VH_loop,-1, 0));
     VH_u  = if(udom, VH_u, 0);
     VH_d  = if(dchan and not IsChannel, shift0(VH_loop*FractExc, 1, 0),
       shift0(VH_loop, 1, 0));
     VH_d  = if(ddom, VH_d, 0);
     VH_u = if(Inlet,  V_inlt*H_inlt, VH_u);
     VH_d = if(Outlet, VH_loop, VH_d);
     # grad(UH), grad(VH)
     dUH\dx = (UH_r - UH_l) / dx;
     dVH\dy = (VH_u - VH_d) / dy;

     # rate of d.f. thickness variation (m.s-1); dH\dT > 0 net gain of material
     dH\dT = -dUH\dx - dVH\dy;
     dH\dT = if(H_loop eq 0 and dH\dT lt 0, 0, dH\dT);

     # update body height; Lax numerical scheme
     H_loop_a = if(dH\dT eq 0, H_loop, 1/8 * (4*H_loop+H_l+H_r+H_u+H_d));
     H_loop = H_loop_a + dT/Loops * dH\dT;
     # correct for mass balance errors introduced by the numerical scheme
     Vol_actual = maptotal(H_loop*CA);
     Vol_exp_loop = Vol_exp_loop + Incr_volexploop;
     MassBalError = if(Vol_exp_loop gt 0, Vol_actual/Vol_exp_loop, 1);
     H_loop = if(MassBalError gt 0, H_loop/MassBalError, H_loop);


     ### 3. NUMERICAL STABILITY ANALYSIS
     # solution is stable if CFL (Courant-Friedrichs-Levy) condition < 1

     # CFL
     Vel = if(H_loop gt 0, max(abs(UH_loop)/H_loop, abs(VH_loop)/H_loop), 0);
     CFL = dT/Loops * Sqrt2 * mapmaximum(Vel) / CL;
     CFL_inlt = dT/Loops * Sqrt2 * abs(V_inlt) / CL;
     CFLmax = max(CFLmax, CFL, CFL_inlt);

     # update number of internal loops
     n = n + 1;
     Unstable = if(CFLmax ge 0.8,            # TRUE if the solution becomes
       boolean(1), boolean(0));              # unstable, FALSE if stable
     n = if(Unstable, n-1, n);               # prevents exiting if n == Loops
   } until n eq Loops or Unstable;           # exit if loops completed or
                                             # solution is unstable

   Loops = if (Unstable,                     # try with more loops if unstable
     roundoff(Loops*1.5), Loops);
#   Loops = if (CFLmax le 0.5, max(roundoff   # ... or w less loops if CFL<0.5
#     (Loops/1.5), MinNLoops), Loops);
#   Unstable = if(CFLmax le 0.5 and Loops eq MinNLoops, boolean(0), Unstable);

 } until not Unstable or Loops gt MaxNLoops; # exit the main loop if not
                                             # unstable or excessive n loops

 # (some dummy identities to avoid getting an exception error due to a PCRaster bug!
 # don't remove; for more info read:
 # http://pcraster.geo.uu.nl/documentation/manual_updates/repeatUntil.html)
 H = H;
 UH = UH;
 VH = VH;
 H_inlt = H_inlt;
 V_inlt = V_inlt;
 Vol_expect = Vol_expect;
 Incr_volexploop = Incr_volexploop;
 Unstable = Unstable;


 ### 4. UPDATE VARIABLES FOR TIMESTEP MARCHING
 
 H  = H_loop;
 UH = UH_loop;
 VH = VH_loop;
 Vol_expect = Vol_exp_loop;
 Loops = if(CFLmax le 0.5, max(roundoff      # don't let CFL drop below 0.5 to
   (Loops/1.5), MinNLoops), Loops);          # control numerical dispersion


 ### F. MODEL OUTPUTS

 # maps at endtime
 #report (end) ${1}_hh = if(H gt 0, H);        # body thickness (m)
 #report (end) ${1}_vv = Vel;                  # velocity last loop (m.s-1)

 # map series
 report (rep_a) ${1}_h = if(H gt 0, H);       # body thickness, no zeros (m)
 #report (rep_a) ${1}_h = H;                   # body thickness (m)
 #report (rep_a) ${1}_zh = Z + 1*H;            # free surface (w vertical exag.)
 #report (rep_a) ${1}_zh = Z - Channels + if(H gt 0, H); 
 #report (rep_a) ${1}_v = Vel;                 # velocity last loop (m.s-1)
 #report (rep_a) ${1}_uu = UH / H;
 #report (rep_a) ${1}_vv = VH / H;
 #report (rep_a) ${1}_uh = UH;
 #report (rep_a) ${1}_vh = VH;
 #report (rep_a) ${1}_kx = K_x;
 #report (rep_a) ${1}_ky = K_y;
 #report (rep_a) ${1}_gp = sqrt(GP_x*GP_x + GP_y*GP_y);
 #report (rep_a) ${1}_i  = sqrt(I_x*I_x + I_y*I_y);
 #report (rep_a) ${1}_t = sqrt(T_x*T_x + T_y*T_y);
 #report (rep_b) ${1}_h = if(H gt 0, H);
 #report (rep_b) ${1}_v = Vel;          
 #report (rep_b) ${1}_gp = sqrt(GP_x*GP_x + GP_y*GP_y);
 #report (rep_b) ${1}_i  = sqrt(I_x*I_x + I_y*I_y);
 #report (rep_b) ${1}_t = sqrt(T_x*T_x + T_y*T_y);

 #time series
 report ${1}_maxh.tss = mapmaximum(H);        # maximum d.f. velocity (m) 
 report ${1}_maxv.tss = mapmaximum(Vel);      # maximum d.f. velocity (m.s-1)
 report ${1}_maxvh.tss = mapmaximum(Vel * H); # maximum d.f. momentum (m.m.s-1)
 report ${1}_cfl.tss = CFLmax;                # maximum Courant condition value
 report ${1}_loops.tss = n;                   # n intern loops whithin timestep
 report ${1}_mass.tss = Vol_actual;           # total mass (m3)
 #report ${1}_massexp.tss = Vol_expect;
 report ${1}_errormass.tss = MassBalError-1;  # mass error (m3.m-3)
 #report ${1}_H.tss = timeoutput(SamplePnts,H);# d.f. thickness at sample points
 #report ${1}_V.tss = timeoutput(SamplePnts,   # d.f. velocity at sample points
 #  Vel);
 #report ${1}_frc.tss = timeoutput(SamplePnts, # force exerted by the flow
 # if(H gt verysmall, Gamma*H*CL *             # against a rigid obstacle (kN)
 # (CL*CosAlfa_y + VH*VH/(2*g*H*H)), 0));
 #report ${1}_inlth.tss = mapmaximum(if(Inlet, H));
 report ${1}_inltvh.tss = mapmaximum(if(Inlet, VH));

 # model params
 report ${1}_${2}_${3}_${4}_${5}_${6}_${7}.tss = # just the simulation params.
   boolean(0);
